/*
  Q Light Controller
  uDMXout.cpp

  Copyright (C) 2007 Heikki Junnila
                     Lutz (ilLUTZminator)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  Version 2 as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details. The license is
  in the file "COPYING".

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "uDMXout.h"
#include "configure_uDMXout.h"
#include "../common/filehandler.h"

#include <stdlib.h>
#include <stdarg.h>

#include <sys/signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <qthread.h>
#include <qstring.h>
#include <qpoint.h>
#include <qpopupmenu.h>
#include <qfile.h>
#include <assert.h>
#include <errno.h>
#include <linux/errno.h>

#define CONF_FILE "uDMXout.conf"

#define ID_CONFIGURE      10
#define ID_ACTIVATE       20

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

static QMutex _mutex;


//
// Exported functions
//
extern "C" OutputPlugin* create(t_plugin_id id)
{
  return new uDMXOut(id);
}

extern "C" void destroy(OutputPlugin* object)
{
  delete object;
}

//
// Class implementation
//
uDMXOut::uDMXOut(t_plugin_id id) : OutputPlugin(id)
{
  for(int n = 0; n < MAXINTERFACES; n++)
    m_device[n] = NULL ;

  m_name          = QString("uDMX Output");
  m_type          = OutputType;
  m_version       = 0x00000001;
  m_deviceName    = QString("uDMX");
  m_firstUniverse = 0;
  m_numOfDevices  = 0;
  m_configDir     = QString("~/.qlc/");

  m_Channels      = 9 ;       // number of channels
  m_DontSendSame  = 0 ;       // Dont send same value more than once -> reduces timing !
  m_Debug         = 0 ;       // NO debug to console
  m_LastInvalidDevice = -1 ;

  // initialize values
  memset(m_values, 0, KChannelMax) ;

  // initialize USB
  usb_set_debug(1);
  usb_init();
}

uDMXOut::~uDMXOut()
{
}

int uDMXOut::usbStringGet(usb_dev_handle *dev, int index, 
                               int langid, char *buf, int buflen)
{
  char    buffer[256];
  int     rval, i;

  if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, 
                             (USB_DT_STRING << 8) + index, langid, 
                             buffer, sizeof(buffer), 1000)) < 0)
    return rval;

  if(buffer[1] != USB_DT_STRING)
    return 0;

  if((unsigned char)buffer[0] < rval)
    rval = (unsigned char)buffer[0];

  rval /= 2;
  /* lossy conversion to ISO Latin1 */
  for(i=1;i<rval;i++)
  {
    if(i > buflen)  /* destination buffer overflow */
      break;
    buf[i-1] = buffer[2 * i];
    if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
      buf[i-1] = '?';
  }
  buf[i-1] = 0;
  return i-1;
}

/* Attempt to open serial device */
int uDMXOut::open()
{
  char    cCmd[8] ;
  QString txt ;
  int     n = m_firstUniverse ;

  m_numOfDevices = 0 ;

  struct usb_bus      *bus;
  struct usb_device   *dev;
  usb_dev_handle      *handle = 0;

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  usb_find_busses();
  usb_find_devices();

  /* Search for uDMX device on all busses */
  for(bus = usb_busses; bus; bus = bus->next)
  {
    for(dev = bus->devices; dev; dev = dev->next)
    {
      if((dev->descriptor.idVendor  == USBDEV_SHARED_VENDOR) && 
         (dev->descriptor.idProduct == USBDEV_SHARED_PRODUCT))
      {
        char    string[256];
        int     len;

        Debug(DEBUG_INFO, "%s %d %s -> Shared Vendor found", __FILE__, __LINE__, __FUNCTION__) ;

        /* we need to open the device in order to query strings */
        handle = usb_open(dev);
        if(!handle)
        {
          Debug(DEBUG_ERROR, "%s %d %s -> ERROR: cannot open USB device: %s",
                __FILE__, __LINE__, __FUNCTION__, usb_strerror()) ;
          continue;
        }

        /* now find out whether the device is uDMX */
        len = usbStringGet(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
        if(len < 0)
        {
          Debug(DEBUG_ERROR, "%s %d %s -> ERROR: cannot query manufacturer for device: %s",
                __FILE__, __LINE__, __FUNCTION__, usb_strerror()) ;
          goto skipDevice;
        }

        Debug(DEBUG_INFO, "%s %d %s -> got Descriptor.iManu [%s]",
              __FILE__, __LINE__, __FUNCTION__, string) ;

        if(strcmp(string, "www.anyma.ch") != 0)
        {
          goto skipDevice;
        }

        len = usbStringGet(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
        if(len < 0)
        {
          Debug(DEBUG_ERROR, "%s %d %s -> ERROR: cannot query product for device: %s",
                __FILE__, __LINE__, __FUNCTION__, usb_strerror()) ;
          goto skipDevice;
        }

        Debug(DEBUG_INFO, "%s %d %s -> got Descriptor.iProd [%s]",
              __FILE__, __LINE__, __FUNCTION__, string) ;

        if(strcmp(string, "uDMX") != 0)
        {
skipDevice:
          usb_close(handle) ;
          handle = NULL ;
        }
        else
        {
          // uDMX device found !
          // store handle in array
          m_device[n] = handle ;

          m_numOfDevices++ ;
          n++ ;

          if ((m_numOfDevices >= MAXINTERFACES) ||
              (n >= (MAXINTERFACES - 1) ))
          {
            goto LoopEnd ;
          }
        }
      }
    }
  }

LoopEnd:

  if(!m_numOfDevices)
  {
    Debug(DEBUG_ERROR, "%s %d %s() ERROR: no valid uDMX devices could be opened ! uDMX output is NOT available !", 
          __FILE__, __LINE__, __FUNCTION__);
  }
  else 
  {
    for(n = m_firstUniverse; n < (m_firstUniverse + m_numOfDevices); n++)
    {
      Debug(DEBUG_INFO, "%s %d %s() uDMX available for universe %d", __FILE__, __LINE__, __FUNCTION__, n) ;

      //writeRange(n * 512, m_values, m_Channels);
      writeChannel((n * 512) + m_Channels, 0);
    }
  }

  Debug(DEBUG_INFO, "%s %d %s() FirstUniverse %d, NumDevs %d", 
        __FILE__, __LINE__, __FUNCTION__, m_firstUniverse, m_numOfDevices ) ;

  return errno ;
}

int uDMXOut::close()
{
  int n;

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  for(n = 0; n < MAXINTERFACES; n++)
  {
    if (m_device[n])
    {
      usb_close(m_device[n]) ;
      m_device[n] = NULL ;
    }
  }

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  return 0;
}

bool uDMXOut::isOpen()
{
  return (m_device[0] != NULL) ;
}

int uDMXOut::configure()
{
  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  Configure_uDMXOut* conf = new Configure_uDMXOut(this);

  if (conf->exec() == QDialog::Accepted)
  {
    m_deviceName    = conf->device();

    // not supported yet
    //m_firstUniverse = conf->firstUniverse();
    m_firstUniverse = 0 ;

    m_Channels      = conf->Channels() ;
    m_DontSendSame  = conf->DontSendSame() ;
    m_Debug         = conf->Debug() ;

    saveSettings() ;
  }

  delete conf;

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  return 0;
}

QString uDMXOut::infoText()
{
  QString t;
  QString str = QString::null;
  str += QString("<HTML><HEAD><TITLE>uDMX Plugin Info</TITLE></HEAD><BODY>");
  str += QString("<TABLE COLS=\"1\" WIDTH=\"100%\"><TR>");
  str += QString("<TD BGCOLOR=\"black\"><FONT COLOR=\"white\" SIZE=\"5\">");
  str += name() + QString("</FONT></TD></TR></TABLE>");
  str += QString("<TABLE COLS=\"2\" WIDTH=\"100%\">");
  str += QString("<TR>\n");
  str += QString("<TD><B>Version</B></TD>");
  str += QString("<TD>");
  t.setNum((version() >> 16) & 0xff);
  str += t + QString(".");
  t.setNum((version() >> 8) & 0xff);
  str += t + QString(".");
  t.setNum(version() & 0xff);
  str += t + QString("</TD>");
  str += QString("</TR>");

  str += QString("<TR>\n");
  str += QString("<TD><B>Status</B></TD>");
  str += QString("<TD>");
  if (isOpen() == true)
  {
    str += QString("<I>Active</I></TD>");
  }
  else
  {
    str += QString("Not Active</TD>");
  }
  str += QString("</TR>");

  str += QString("</TR>");
  str += QString("</TABLE>");

  str += QString("</BODY></HTML>");

  return str;
}

int uDMXOut::setConfigDirectory(QString dir)
{
  m_configDir = dir;
  return 0;
}

int uDMXOut::saveSettings()
{
  QString s;
  QString t;

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  QString fileName = m_configDir + QString(CONF_FILE);
  Debug(DEBUG_INFO, "%s %d %s: Saving config to %s", __FILE__, __LINE__, __FUNCTION__, fileName.data() ) ;
  QFile file(fileName);

  if (file.open(IO_WriteOnly))
  {
    // Comment
    s = QString("# uDMXout Plugin Configuration\n");
    file.writeBlock((const char*) s, s.length());

    // Entry type
    s = QString("Entry = ") + name() + QString("\n");
    file.writeBlock((const char*) s, s.length());

    // Device (/dev/ttyS)
    s = QString("Device = ") + m_deviceName + QString("\n");
    file.writeBlock((const char*) s, s.length());

    // 
    t.sprintf("%d",m_firstUniverse);
    s = QString("FirstUniverse = ") + t + QString("\n");
    file.writeBlock((const char*) s, s.length());

    // channels
    t.sprintf("Channels = %d\n", m_Channels) ;
    file.writeBlock((const char*) t, t.length()) ;

    // Dont send same value more than once
    t.sprintf("DontSendSame = %d\n", m_DontSendSame) ;
    file.writeBlock((const char*) t, t.length()) ;

    // debug
    t.sprintf("Debug = %d\n", m_Debug) ;
    file.writeBlock((const char*) t, t.length()) ;

    file.close();
  }
  else
  {
    perror("file.open");
    Debug(DEBUG_ERROR, "%s %d %s() ERROR: Unable to save uDMXout configuration",
           __FILE__, __LINE__, __FUNCTION__);
  }

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  return errno;
}

int uDMXOut::loadSettings()
{
  QString fileName;
  QPtrList <QString> list;

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  fileName = m_configDir + QString(CONF_FILE);

  FileHandler::readFileToList(fileName, list);

  for (QString* s = list.first(); s != NULL; s = list.next())
  {
    if (*s == QString("Entry"))
	  {
	    s = list.next();
	    if (*s == name())
	    {
	      createContents(list);
	    }
	  }
  }

  Debug(DEBUG_ALL, "%s %d %s", __FILE__, __LINE__, __FUNCTION__) ;

  return 0;
}

void uDMXOut::createContents(QPtrList <QString> &list)
{
  for (QString* s = list.next(); s != NULL; s = list.next())
  {
    if (*s == QString("Entry"))
	  {
	    s = list.prev();
	    break;
	  }
    else if (*s == QString("Device"))
	  {
	    m_deviceName = *(list.next());
	  }
    else if (*s == QString("FirstUniverse"))
	  {
	    m_firstUniverse = list.next()->toInt() ;
      m_firstUniverse = MIN(m_firstUniverse, MAXINTERFACES - 1) ;
	  }
    else if (*s == QString("Channels"))
    {
      m_Channels = list.next()->toInt() ;
    }
    else if (*s == QString("DontSendSame"))
    {
      m_DontSendSame = list.next()->toInt() ;
    }
    else if (*s == QString("Debug"))
    {
      m_Debug = list.next()->toInt() ;
    }
  }
}

void uDMXOut::contextMenu(QPoint pos)
{
  QPopupMenu* menu = new QPopupMenu();
  menu->insertItem("Configure...", ID_CONFIGURE);
  menu->insertSeparator();
  menu->insertItem("Activate", ID_ACTIVATE);

  connect(menu, SIGNAL(activated(int)),
          this, SLOT(slotContextMenuCallback(int)));
  menu->exec(pos, 0);
  delete menu;
}

void uDMXOut::slotContextMenuCallback(int item)
{
  switch(item)
  {
  case ID_CONFIGURE:
   configure();
    break;

  case ID_ACTIVATE:
    activate();
    break;

  default:
    break;
  }
}

void uDMXOut::activate()
{
  emit activated(this);
}

//
// Write a value to a channel
//
int uDMXOut::writeChannel(t_channel channel, t_value value)
{
  // Make it simple an actually use only ONE function
  // no difference between writeChannel and writeRange here ...
  //m_values[channel] = value ;
  //return ( writeRange(channel, &m_values[channel], 1) ) ;

  int           nBytes ;
  unsigned char buffer[2];

  Debug(DEBUG_ALL, "%s %d %s(address %4d[0x%04X], value %3d)", 
        __FILE__, __LINE__, __FUNCTION__, channel, channel, value) ;


  if ((m_DontSendSame) && (m_values[channel] == value))
  {
    Debug(DEBUG_INFO, "%s %d %s(Channel %d -> Value %d did not change)", 
          __FILE__, __LINE__, __FUNCTION__, channel, value) ;
    return true ;
  }
  else
    m_values[channel] = value ;

  // which one is the universe to write to?
  int device       = int(channel / 512) ;
  if (!m_device[device])
  {
    Debug(DEBUG_WARN, "%s %d %s(device %d not opened)", 
          __FILE__, __LINE__, __FUNCTION__, device) ;
    return false ;
  }

  //what is the first channel on that device?
  // eg. address is 512, it would be channel 0 on device 1
  int firstChannel = int(channel % 512) ;

  if (firstChannel > m_Channels)
  {
    Debug(DEBUG_WARN, "%s %d %s(Channel %d > ChannelMax %d)", 
          __FILE__, __LINE__, __FUNCTION__, firstChannel, m_Channels) ;
    return false ;
  }

  memset(buffer, 0, sizeof(buffer)) ;

  nBytes = usb_control_msg(m_device[device], USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                           cmd_SetSingleChannel, value, firstChannel, (char *)buffer, sizeof(buffer), 50000);
  if(nBytes < 0)
  {
    Debug(DEBUG_ERROR, "%s %d %s() USB-ERROR: %s",
          __FILE__, __LINE__, __FUNCTION__, usb_strerror() ) ;
  }
  return true ;
}

//
// Write num values starting from address
//
int uDMXOut::writeRange(t_channel address, t_value* values, t_channel num)
{
  assert(values) ;

  int               r = 0 ;
  QString           txt ;
  unsigned char     buf[16] ;
  t_value *         pValue ;
  int               iChannelsToWrite, iBlockSize, iDevToWrite ;

  _mutex.lock();

  // store memory for output
  pValue = m_values + address ;
  memcpy(pValue, values, num * sizeof(t_value));

  // which one is the first universe to write to?
  int device       = int(address / 512) ;

  //what is the first channel on that device?
  // eg. address is 512, it would be channel 0 on device 1
  int firstChannel = address % 512 ;


  //write the first chunk
  int numToWrite = MIN(512 - firstChannel, num);


  Debug(DEBUG_ALL, "%s %d %s(address %4d[0x%04X], num %4d)", 
        __FILE__, __LINE__, __FUNCTION__, address, address, num) ;

  while(num > 0)
  {
    if(device >= m_numOfDevices)
    {
      // not enough devices ...
      if ((device > m_LastInvalidDevice) ||
          (m_Debug >= DEBUG_ALL))
      {
        m_LastInvalidDevice = device ;
        Debug(DEBUG_WARN, "%s %d %s() ERROR: tried to write to non-existing device %d",
              __FILE__, __LINE__, __FUNCTION__, device) ;
      }
      _mutex.unlock();
      return -1;
    }
    else
    {
      // OK, ready to send
      int nBytes ;

      iChannelsToWrite = MIN(numToWrite, m_Channels)  ;

      while (iChannelsToWrite > 0)
      {
        // size bigger than 256 creates too much "broken pipe" errors
        // so limit size to 256
        iBlockSize  = MIN(iChannelsToWrite, 256) ;

        nBytes = usb_control_msg(m_device[device], USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                cmd_SetChannelRange, iBlockSize, firstChannel,
                                (char *)values, iBlockSize, 50000);
        Debug(DEBUG_INFO, "%s %d %s(): %i bytes returned", __FILE__, __LINE__, __FUNCTION__, nBytes) ;
        if(nBytes < 0)
        {
          Debug(DEBUG_ERROR, "%s %d %s() USB-ERROR: %s",
                __FILE__, __LINE__, __FUNCTION__, usb_strerror() ) ;
        }

        iChannelsToWrite  = iChannelsToWrite - (iBlockSize) ;
        firstChannel      = firstChannel     + (iBlockSize) ;
        pValue            = pValue           + (iBlockSize) ;
      }

      firstChannel = 0;
      device ++;
      num -= numToWrite;
      values += numToWrite;
      if(num >= 512)
        numToWrite = 512;
      else
        numToWrite = num;
    }
  }

  _mutex.unlock();
  return r;
}

// Needed for all plugins, for example to show DMX monitor
int uDMXOut::readChannel(t_channel channel, t_value &value)
{
  _mutex.lock();

  value = m_values[channel] ;

  _mutex.unlock();

  return 0;
}

int uDMXOut::readRange(t_channel address, t_value* values, t_channel num)
{
  _mutex.lock();

  for (t_channel i = 0; i < num; i++)
  {
    values[i] = m_values[i];
  }

  _mutex.unlock();

  return 0;
}

// debugging: makes call for debug shorter ...
bool uDMXOut::Debug(int debuglevel, char * format, ...)
{
  // That size should be big enough
  // little security risk of buffer overrun but another way
  // is too complicated ...
  char string[4096] ;

  if (m_Debug < debuglevel)
  {
    return true ;
  }

  va_list argzeiger ;
  va_start(argzeiger, format) ;
  vsprintf(string, format, argzeiger) ;
  va_end(argzeiger) ;

  qDebug(string) ;

  return true ;
}
